#include "ruby.h"
#include "intern.h"
#include "env.h"
#include "internal.h"
#include "state.h"
#include "runtime.h"
/*
 * call-seq: Ruby::Frame.new
 *
 * Creates a new frame object. A frame object
 * represents a location that code was run.
 */
VALUE rb_frame_init(VALUE self) {
    rb_iv_set(self, "@self", Qnil);
    rb_iv_set(self, "@current_method", Qnil);
    rb_iv_set(self, "@original_method", Qnil);
    rb_iv_set(self, "@file", rb_str_new2("(unknown)"));
    rb_iv_set(self, "@line", INT2FIX(0));
    rb_iv_set(self, "@current_class", Qnil);
    rb_iv_set(self, "@metadata", Qnil);
    return Qnil;
}


VALUE rb_wrap_frame(struct FRAME *fr, int use_prev) {
    VALUE obj;
    ID func;

    //printf("wrapping %p / %p\n", fr, fr->prev);
    /*
     * Note the use of fr->prev here. Thats because
     * a new frame is allocated to call this method,
     * so we actually want the one right behind our own
     * frame. */
    use_prev = 1;
    
    obj = rb_class_new_instance(0, NULL, CPATH(Ruby::Frame));
    rb_iv_set(obj, "@self", fr->prev->self);
    if(fr->prev && fr->prev->last_func) {
        func = use_prev ? fr->prev->last_func : fr->last_func;
        rb_iv_set(obj, "@current_method", ID2SYM(func));
    } else {
        rb_iv_set(obj, "@current_method", Qnil);
    }
    if(fr->prev && fr->prev->orig_func) {
        func = use_prev ? fr->prev->orig_func : fr->orig_func;
        rb_iv_set(obj, "@original_method", ID2SYM(fr->prev->orig_func));
    } else {
        rb_iv_set(obj, "@original_method", Qnil);
    }

    rb_iv_set(obj, "@file", rb_str_new2(fr->node->nd_file));
    rb_iv_set(obj, "@line", INT2NUM(nd_line(fr->node)));
    if(fr->prev->last_class) {
        rb_iv_set(obj, "@current_class", fr->prev->last_class);
    } else {
        rb_iv_set(obj, "@current_class", Qnil);
    }

    if(!fr->prev->metadata || !RTEST(fr->prev->metadata)) {
        fr->prev->metadata = rb_hash_new();
    } 

    rb_iv_set(obj, "@metadata", fr->prev->metadata);

    return obj;
}

/* call-seq: Ruby::Frame.current
 *
 * Returns a frame representing the current location.
 */
VALUE rb_frame_current(VALUE self) {
    return rb_wrap_frame(ruby_frame, 1);
}

/*
 * call-seq: frame.position
 *
 * Returns an array containing the file and line
 * the frame represents.
 */
VALUE rb_frame_position(VALUE self) {
    VALUE ary;

    ary = rb_ary_new3(2, 
            rb_iv_get(self, "@file"),
            rb_iv_get(self, "@line"));

    // ary = rb_ary_new3(2, rb_str_new2(fr->node->nd_file), FIX2INT(nd_line(fr->node)));

    return ary;
}

/*
 * call-seq: frame.nice_string
 *
 * Returns a string representation of the frame.
 * IE: blah.rb:8:in 'do_it'
 */
VALUE rb_frame_nice_string(VALUE self) {
    VALUE str, last_meth;

    str = rb_str_new3(rb_iv_get(self, "@file"));
    rb_str_cat(str, ":", 1);
    rb_str_append(str, rb_fix2str(rb_iv_get(self, "@line"), 10));
    last_meth = rb_iv_get(self, "@current_method");
    if(RTEST(last_meth)) {
        rb_str_cat(str,":in '", 5);
        rb_str_append(str, rb_str_new2(rb_id2name(SYM2ID(last_meth))));
        rb_str_cat(str,"'",1);
    }

    return str;
}

/*
 * call-seq: Ruby::Backtrace.new
 *
 * Creates a new backtrace object.
 */
VALUE rb_backtrace_init(VALUE self) {
    rb_iv_set(self, "@frames", rb_ary_new());
    rb_iv_set(self, "@file", rb_str_new2("(unknown)"));
    rb_iv_set(self, "@line", INT2FIX(0));
    rb_iv_set(self, "@current_method", Qnil);
    return Qnil;
}

/*
 * call-seq bt.import_position
 *
 * Imports the current position into the backtrace
 * as the origin location. The current position consists
 * of the current file (@file), the current line (@line),
 * and the current method (@current_method).
 */
VALUE rb_backtrace_import_position(VALUE self) {
    rb_iv_set(self, "@file", rb_str_new2(ruby_sourcefile));
    rb_iv_set(self, "@line", INT2NUM(ruby_sourceline));
    if(ruby_frame->last_func) {
        rb_iv_set(self, "@current_method", ID2SYM(ruby_frame->last_func));
    }
    return self;
}

/*
 * call-seq: bt.add_frame frame
 *
 * Appends a frame to the backtrace.
 */
VALUE rb_backtrace_add_frame(VALUE self, VALUE frame) {
    rb_ary_push(rb_iv_get(self, "@frames"), frame);
    return Qnil;
}

/*
 * call-seq: bt.print([head, [tail, [max]])
 *
 * Prints the frames of the backtrace. Auto-truncates
 * the output according the values head, tail and max.
 *
 * By default head is 8, tail is 7 and max is 22.
 * These are the same values used by the internal backtrace
 * print function.
 *
 * The defaults can be accessed as class variables @@head_trace,
 * @@tail_trace, and @@max_trace of Ruby::Backtrace.
 */
VALUE rb_backtrace_print(int argc, VALUE *argv, VALUE self) {
    VALUE frames, head, tail, max, str, output;
    int i, skipped;
    
    frames = rb_iv_get(self, "@frames");
    if(NIL_P(frames)) return Qnil;

    head = Qnil;
    tail = Qnil;
    max = Qnil;
   
    rb_scan_args(argc, argv, "03", &head, &tail, &max);

    if(!RTEST(head)) {
        head = rb_cv_get(CLASS_OF(self), "@@head_trace");
    }

    if(!RTEST(tail)) {
        tail = rb_cv_get(CLASS_OF(self), "@@tail_trace");
    }

    if(!RTEST(max)) {
        max = rb_cv_get(CLASS_OF(self), "@@max_trace");
    }

    output = rb_str_new(NULL, 0);
    skipped = RARRAY(frames)->len - FIX2INT(head) - FIX2INT(tail);

    for(i = 0; i < RARRAY(frames)->len; i++) {
        str = rb_frame_nice_string(RARRAY(frames)->ptr[i]);
        rb_str_cat(output, "\tfrom ", 6);
        rb_str_append(output, str);
        rb_str_cat(output, "\n", 1);

        if(i == (FIX2INT(head) - 1) && RARRAY(frames)->len > FIX2INT(max)) {
            rb_str_cat(output, "\t ... ", 6);
            rb_str_append(output, rb_fix2str(INT2FIX(skipped), 10));
            rb_str_cat(output, " levels...\n",11);
            i = RARRAY(frames)->len - FIX2INT(tail) - 1;
        }
    }

    return output;
}

/*
 * call-seq: Ruby.error_handler { p $! }
 *
 * When an error needs to be emitted to the user, the
 * closure is called. $! should be used to obtain
 * the last exception called (typically the reason
 * the error was emitted.
 */
VALUE rb_set_error_handler(VALUE self) {
    ruby_errhandler = rb_block_proc();
    return Qnil;
}

/*
 * call-seq: bt.location
 *
 * Returns a string that represents the location of the
 * backtrace.
 */
VALUE rb_backtrace_location(VALUE self) {
    VALUE r, str;
    char *i;

    r = rb_iv_get(self, "@file");
    str = rb_str_new3(r);
    
    r = rb_iv_get(self, "@line");
    if(FIX2INT(r) > 0) {
        rb_str_cat(str, ":", 1);
        rb_str_append(str, rb_fix2str(r, 10));
    }
    r = rb_iv_get(self, "@current_method");
    if(RTEST(r)) {
        i = rb_id2name(SYM2ID(r));
        if(i != NULL) {
            rb_str_cat(str, " in '", 5);
            rb_str_cat2(str, i);
            rb_str_cat(str, "'", 1);
        }
    }
    return str;
}

void Init_frame() {
    VALUE cFrame, cBacktrace;
    cFrame = rb_define_class_under(CPATH(Ruby), "Frame" , rb_cObject);
    cBacktrace = rb_define_class_under(CPATH(Ruby), "Backtrace", rb_cObject);

    rb_define_singleton_method(cFrame, "current", rb_frame_current, 0);
    rb_define_method(cFrame, "position", rb_frame_position, 0);
    rb_attr(cFrame, rb_intern("self"), 1, 1, 0);
    rb_attr(cFrame, rb_intern("file"), 1, 1, 0);
    rb_attr(cFrame, rb_intern("line"), 1, 1, 0);
    rb_attr(cFrame, rb_intern("current_method"), 1, 1, 0);
    rb_attr(cFrame, rb_intern("original_method"), 1, 1, 0);
    rb_attr(cFrame, rb_intern("current_class"), 1, 1, 0);
    rb_attr(cFrame, rb_intern("metadata"), 1, 0, 0);
    rb_define_method(cFrame, "nice_string", rb_frame_nice_string, 0);
    rb_define_method(cFrame, "to_str", rb_frame_nice_string, 0);
    rb_define_method(cFrame, "to_s", rb_frame_nice_string, 0);
    
    rb_attr(cBacktrace, rb_intern("frames"), 1, 1, 0);
    rb_attr(cBacktrace, rb_intern("file"), 1, 1, 0);
    rb_attr(cBacktrace, rb_intern("line"), 1, 1, 0);
    rb_attr(cBacktrace, rb_intern("current_method"), 1, 1, 0);
    rb_define_method(cBacktrace, "initialize", rb_backtrace_init, 0);
    rb_define_method(cBacktrace, "location", rb_backtrace_location, 0);
    rb_define_method(cBacktrace, "print", rb_backtrace_print, -1);
    rb_define_method(cBacktrace, "add_frame", rb_backtrace_add_frame, 1);
    rb_cv_set(cBacktrace, "@@head_trace", INT2FIX(7));
    rb_cv_set(cBacktrace, "@@tail_trace", INT2FIX(5));
    rb_cv_set(cBacktrace, "@@max_trace", INT2FIX(22));

    rb_define_singleton_method(CPATH(Ruby), "error_handler", rb_set_error_handler, 0);
}
