class Thread::OS

    # Creates a new Thread::OS, running the given block
    # in the thread over and over again,
    def self.daemon(&block)
        new do
            loop(&block)
        end
    end

    # Kills off all OS threads except the calling
    # one, isolating it.
    def self.isolate
        Thread::OS.list.each do |th|
            next if th == Thread::OS.current
            th.cancel
        end

        Thread::OS.yield while Thread::OS.list.size != 1
    end

    class Mutex

        # Locks the mutex, runs the given block, and
        # ensures that the mutex is unlocked after
        # the block exits.
        def synchronize
            lock
            begin
                yield
            ensure
                unlock
            end
        end
    end

    # An error checking mutex that keeps track of the owner 
    # of the mutex. If the current owner attempts to relock
    # the mutex when it is already locked, it just returns false
    # without blocking.
    class RecursiveMutex < Mutex
        def initialize
            @owner = nil
            super
        end
        
        # Locks the mutex. If the caller is the current owner,
        # returns false. Otherwise blocks the calling
        # thread until the mutex is locked.
        def lock
            if @owner == Thread::OS.current
                return false
            end

            super
            @owner = Thread::OS.current

            return true
        end

        # Trys to lock the mutex. If the caller is the current owner,
        # returns false. Otherwise performs the same as +Thread::OS::Mutex+.
        def trylock
            if @owner == Thread::OS.current
                return false
            end

            super
            @owner = Thread::OS.current

            return true
        end

        # Resets the owner and unlocks the mutex.
        def unlock
            @owner = nil
            super
        end

        # Returns the current owner of the mutex.
        def owner
            @owner
        end
    end
end
