#
# An "addon" to the Rubicon testsuite to test for "missing" tests of
# a class/module under test.
#

# Rubicon::MethodChecker is used as a "namespace" to avoid cluttering
# the Rubicon::TestCase class. It is used from that class (see below).

class Rubicon::MethodChecker

  @@alias_to_op = {
    "PLUS"       => '+',
    "MINUS"      => '-',

    "UPLUS"      => "+@",
    "UMINUS"     => "-@",

    "MUL"        => '*',
    "DIV"        => "/",
    "MOD"        => "%",
    "POW"        => "**",

    "AND"        => '&',
    "OR"         => '|',
    "XOR"        => '^',
    "REV"        => "~",

    "LSHIFT"     => '<<',
    "RSHIFT"     => ">>",

    "CMP"        => '<=>',
    "EQUAL"      => '==',
    "VERY_EQUAL" => '===',

    "AREF"       => '[]',
    "ASET"       => '[]=',

    "LT"         => '<',
    "LE"         => '<=',
    "GT"         => '>',
    "GE"         => '>=',

    "MATCH"      => '=~',
    "BACKTICK"   => '`',
  }
  @@op_to_alias = @@alias_to_op.invert

  def initialize(test_obj)
    @test_obj = test_obj
  end

  def methods_to_test(inherited=false)
    names = tested_class.public_instance_methods(inherited)
    instance_display_names(names)
  end
  
  def singleton_methods_to_test
    #names = tested_class.public_methods(false)
    names = tested_class.singleton_methods(false)
    singleton_display_names(names)
  end

  def tested_methods
    names = test_class.public_instance_methods(false).select do |m|
      m =~ /^test_[a-zA-Z]/ && m !~ /^test_s_/
    end.map do |m|
      m.sub(/^test_/, "")
    end.map do |m|
      @@alias_to_op[m] || m
    end
    instance_display_names(names)
  end
  
  def tested_singleton_methods
    names = test_class.public_instance_methods(false).select do |m|
      m =~ /^test_s_[a-zA-Z]/
    end.map do |m|
      m.sub(/^test_s_/, "")
    end.map do |m|
      @@alias_to_op[m] || m
    end
    singleton_display_names(names)
  end
  
  def display_names(prefix, names)
    names.map do |name|
      if @@op_to_alias[name]
        prefix + @@op_to_alias[name] + " ( operator " + name + " )"
      else
        #if name !~ /^[a-zA-Z0-9_]+([?!=])?$/
        #  $stderr.puts "strange method: #{name}"
        #  exit!(1)
        #end
        prefix + name
      end
    end
  end

  def instance_display_names(names)
    display_names("test_", names)
  end

  def singleton_display_names(names)
    display_names("test_s_", names)
  end

  def test_class
    case @test_obj.class.name
    when "TestArray"   ; ArrayBase
    when "TestHash"    ; HashBase
    when "TestString"  ; StringBase
    else
      @test_obj.class
    end
  end

  def tested_class
    klass = @test_obj.class
    Object.const_get(klass.to_s.sub(/^Test/, ''))
  end

  def class_defined?
    klass = @test_obj.class
    if klass == Rubicon::TestCase
      false
    else
      Object.const_defined?(klass.to_s.sub(/^Test/, ''))
    end
  end

end

#----------------------------------------------------------------------

class Rubicon::TestCase

  def test_95_untested_methods
    checker = Rubicon::MethodChecker.new(self)
    return unless checker.class_defined?
    assert_equal([],
		 checker.methods_to_test -
                   checker.tested_methods,
		 "Missing tests in #{checker.test_class}")
  end

  def test_96_untested_class_methods
    checker = Rubicon::MethodChecker.new(self)
    return unless checker.class_defined?
    assert_equal([],
		 checker.singleton_methods_to_test -
                   checker.tested_singleton_methods,
		 "Missing tests in #{checker.test_class} (for singletons)")
  end

  def test_97_unknown_methods_tested
    checker = Rubicon::MethodChecker.new(self)
    return unless checker.class_defined?
    assert_equal([],
		 checker.tested_methods -
                   checker.methods_to_test(! $rubicon_missing_strict),
		 "Unknown tests in #{checker.test_class}")
  end

  def test_98_unknown_class_methods_tested
    checker = Rubicon::MethodChecker.new(self)
    return unless checker.class_defined?
    assert_equal([],
		 checker.tested_singleton_methods -
                   checker.singleton_methods_to_test,
		 "Unknown tests in #{checker.test_class} (for singletons)")
  end

end

