#include "ruby.h"
#include "internal.h"
#include <stdlib.h>
#include <pthread.h>


ruby_state *alloc_ruby_state() {
    int i;
    ruby_state *st = (ruby_state*)malloc(sizeof(ruby_state));
    memset(st, 0, sizeof(ruby_state));
    st->ruby_debug = Qfalse;
    st->ruby_verbose = Qfalse;
    st->sflag = 0;
    st->xflag = 0;
    
    st->initialized = 0;
    st->ruby_in_compile = 0;
    st->gc = (struct gc_state*)malloc(sizeof(struct gc_state));
    st->gc->heap_slots = HEAP_MIN_SLOTS;
    st->gc->malloc_limit = GC_MALLOC_LIMIT;
    i = pthread_mutex_init(&st->gc->lock, NULL);
    if(i) {
        printf("Unable to create GC lock\n");
        abort();
    }

    /* from eval.c */
    st->safe_level = 0;
    st->ruby_running = 0;
    st->frame_unique = 0;
    st->block_unique = 0;
    st->ruby_cref = 0;
    st->current_thread = 0;
    st->trace_func = 0;
    st->tracing = 0;
    
    st->ruby_errinfo = Qnil;
    st->ruby_errhandler = Qnil;
    st->ruby_eval_tree_begin = 0;
    st->ruby_eval_tree = 0;
    st->thread_pending = 0;
    st->thread_init = 0;
    st->trap_immediate = 1;
    st->thread_tick = 500;
    st->prohibit_interrupt = 1;

    st->chdir_blocking = 0;
    st->chdir_thread = Qnil;

    st->frames = st->top_frame = &(st->initial_frame);
    st->ruby_iter = &(st->initial_iter);

    /* dln.c */
    st->dln_init_p = 0;
    st->dln_reloc_tbl = NULL;

    /* hash.c */
    st->path_tainted = -1;

    /* variable.c */

    st->lineno = INT2FIX(0);

    /* ruby.c */
    st->do_loop = Qfalse;
    st->do_print = Qfalse;
    st->do_check = Qfalse;
    st->do_line = Qfalse;
    st->do_split = Qfalse;

    i = pthread_mutex_init(&st->xcritical, NULL);
    if(i) {
        printf("Unable to create critical mutex!\n");
    }

    st->main_osthread = pthread_self();

    st->parent = st;
    return st;
}

void Init_stack(void*);
rb_thread_t rb_thread_alloc(VALUE);

ruby_state *inherit_from_state(ruby_state *state) {
    int top;
    ruby_state *new_state;
    rb_thread_t cur, maint;
    if(state == NULL) {
        state = rb_get_state();
    }

    new_state = (ruby_state *)malloc(sizeof(ruby_state));

    memcpy(new_state, state->pristine, sizeof(ruby_state));
    new_state->parent = state;
    new_state->gc_stack_start = 0;

    rb_set_state(new_state);
    new_state->gc = state->gc;

    Init_stack((void*)&top);
    /* ruby_init(); */
    STATE_VAR(primary_thid) = pthread_self();
    STATE_VAR(main_osthread) = state->main_osthread;
    STATE_VAR(os_threads) = state->os_threads;
    STATE_VAR(prot_tag) = 0;
    STATE_VAR(ruby_scope) = 0;
    STATE_VAR(frames) = 0;
    STATE_VAR(gc) = state->gc;
    pthread_mutex_init(&new_state->xcritical, NULL);
    STATE_VAR(xcritical_owner) = 0;

    maint = rb_thread_alloc(CPATH(Thread));
    cur = maint->prev = maint->next = maint;
    STATE_VAR(main_thread) = maint;
    STATE_VAR(current_thread) = cur;

    if(!state->sub_states) {
        state->sub_states = rb_ary_new();
    }
    rb_ary_push(state->sub_states, Data_Wrap_Struct(CPATH(Ruby::State),0,0,new_state));
    return new_state;
}

ruby_state *rb_dup_state(ruby_state *state) {
    ruby_state *new;
    new = (ruby_state *)malloc(sizeof(ruby_state));
    memcpy(new, state, sizeof(ruby_state));
    return new;
}

static pthread_key_t ruby_state_key;

void init_ruby_key() {
    pthread_key_create(&ruby_state_key, NULL);
}

void init_ruby_state() {
    ruby_state *st = alloc_ruby_state();
    pthread_setspecific(ruby_state_key, st);
    // global_ruby_state = alloc_ruby_state();
}

void rb_set_state(ruby_state *st) {
    pthread_setspecific(ruby_state_key, st);
}

ruby_state *rb_get_state() {
    ruby_state *st = (ruby_state*)(pthread_getspecific(ruby_state_key));
    return st;
}

int rb_lock_xcritical(char *file, int line) {
    int try;
    try = pthread_mutex_trylock(&STATE_VAR(xcritical));
    if(try) {
        // If we're the owner, no need to relock it.
        if(pthread_equal(pthread_self(), STATE_VAR(xcritical_owner))) {
            return 0;
        } else {
            pthread_mutex_lock(&STATE_VAR(xcritical));
        }
        // printf("");
    } 
    STATE_VAR(xcritical_owner) = pthread_self();
    return 1;
}

int rb_unlock_xcritical() {
    int try;

    try = pthread_mutex_unlock(&STATE_VAR(xcritical));

    return try;
}

#define GET_STATE(self) ((ruby_state*)DATA_PTR(self))

void rb_osthread_lock(VALUE);
void rb_osthread_unlock(VALUE);

VALUE rb_state_threads(VALUE self) {
    ruby_state *st;
    VALUE ary;
    rb_thread_t th;

    ary = rb_ary_new();
    st = GET_STATE(self);

    if(!st) {
        printf("Invalid thread state!!\n");
    }

    rb_osthread_lock(st->thread);
#undef main_thread
    th = st->main_thread;
    do {
        switch (th->status) {
            case THREAD_RUNNABLE:
            case THREAD_STOPPED:
            case THREAD_TO_KILL:
                rb_ary_push(ary, th->thread);
            default:
                break;
        }
        if(RARRAY(ary)->len > 5) {
            abort();
        }
        th = th->next;
    } while(th != st->main_thread);

    rb_osthread_unlock(st->thread);
    return ary;
}

VALUE rb_state_heap_size(VALUE self) {
    unsigned long sz;

    sz = GC_VAR(heaps_length) * sizeof(RVALUE) * GC_VAR(heap_slots);

    return UINT2NUM(sz);
}

VALUE rb_state_num_objects(VALUE self) {
    return UINT2NUM(GC_VAR(num_objects));
}

VALUE rb_state_set_switch(int argc, VALUE *argv, VALUE self) {
    if(argc == 0) {
        GET_STATE(self)->on_switch = rb_block_proc();
    } else if(RTEST(argv[0])) {
        GET_STATE(self)->on_switch = argv[0];
    } else {
        GET_STATE(self)->on_switch = 0;
    }
    return Qnil;
}

VALUE rb_state_set_interrupt(int argc, VALUE *argv, VALUE self) {
    if(argc == 0) {
        GET_STATE(self)->interrupt = rb_block_proc();
    } else if(RTEST(argv[0])) {
        GET_STATE(self)->interrupt = argv[0];
    } else {
        GET_STATE(self)->interrupt = 0;
    }
    return Qnil;
}

void rb_state_run_interrupt() {
    int crit = STATE_VAR(thread_critical);
    int proh = STATE_VAR(prohibit_interrupt);
    STATE_VAR(thread_critical) = 1;
    STATE_VAR(prohibit_interrupt) = 1;
    rb_funcall(STATE_VAR(interrupt), _S(call), 0);
    STATE_VAR(thread_critical) = crit;
    STATE_VAR(prohibit_interrupt) = proh;
}

void Init_state() {
    VALUE cState;

    cState = rb_define_class_under(CPATH(Ruby), "State", rb_cObject);

    rb_define_method(cState, "threads", rb_state_threads, 0);
    rb_define_method(cState, "heap_size", rb_state_heap_size, 0);
    rb_define_method(cState, "num_objects", rb_state_num_objects, 0);
    rb_define_method(cState, "on_switch", rb_state_set_switch, -1);
    rb_define_method(cState, "on_interrupt", rb_state_set_interrupt, -1);
}
