require 'stack'
require 'backtrace'
require 'test/unit'
require 'osthread'

class TestOSThread < Test::Unit::TestCase

    def test_current
        assert_kind_of(Thread::OS, Thread::OS.current)
    end

    def test_thread_id
        assert_kind_of(Numeric, Thread::OS.thread_id)
    end

    def test_locals
        val = Object.new
        assert_nothing_raised do
            Thread::OS.current[:name] = val
        end
        assert_equal(val, Thread::OS.current[:name])
    end

    def test_join_exc
        th = Thread::OS.start do
            raise ArgumentError
        end

        assert_raises(ArgumentError) do
            th.join
        end
    end

    def test_join_result
        th = Thread::OS.start do
            0x8989
        end

        assert_equal(0x8989, th.join)
    end

    def test_detach_and_cancel
        th = Thread::OS.new do
            sleep 1000
        end

        assert_nothing_raised do
            th.detach
        end

        assert_raises(Thread::OS::InvalidThread) do
            th.join
        end

        th.cancel
        Thread::OS.yield
        assert(!Thread::OS.list.include?(th))
    end

    def test_deadlock
        assert_raises(Thread::OS::DeadLock) do
            Thread::OS.current.join
        end
    end

    def notest_equal
        th = Thread::OS.new do
            $th = Thread::OS.current
        end
        th.join
        assert_equal(th, $th)
    end

    def test_list
        assert(Thread::OS.list.include?(Thread::OS.current))
    end

    def test_pause
        $g = 0
        th = Thread::OS.new do
            loop {
                $g += 1
                sleep 0.2
            }
        end
        th.pause
        assert_equal("sleep", th.status)
        q = $g
        sleep 1
        assert_equal(q, $g)
    end

    def test_pause2
        th = Thread::OS.new do
            sleep 10
        end

        th.pause
        th.pause
    end

    def notest_isolation
        th = Thread::OS.new do
            sleep 100000
        end
        Thread::OS.isolate
        assert_equal(1, Thread::OS.list.size)
        assert_equal(Thread::OS.current, Thread::OS.list.first)
    end

    def test_daemon
        $g = 0
        th = Thread::OS.daemon do
            $g += 1
            sleep 0.1
        end

        sleep 0.3
        assert($g > 1)
        th.cancel
        th.join
    end

    def test_policy
        pol = Thread::OS.current.policy
    end

    def notest_kill
        $hupped = false
        th = Thread::OS.new do
            Signal.trap("HUP") do
                $hupped = true
            end
            sleep 10000
        end

        th.kill("HUP")

        Thread::OS.yield

        assert_equal(true, $hupped)
    end

    def inc(a,i)
        Thread.new do
            loop do
                a[i] = a[i] + 1
            end
        end
    end

    def test_mxn
        ary = [0, 0, 0, 0]
        th = Thread::OS.new do
            begin
                t1 = inc ary, 0
                t2 = inc ary, 1
                sleep 1000
            rescue RuntimeError
                t1.kill
                t2.kill
            end
        end

        t1 = inc ary, 2
        t2 = inc ary, 3

        sleep 1

        assert(ary[0] > 10)
        assert(ary[1] > 10)
        assert(ary[2] > 10)
        assert(ary[3] > 10)

        th.raise RuntimeError
        t1.kill
        t2.kill
    end

    def test_raise
        th = Thread::OS.new do
            sleep 1000
        end

        mesg = "<< die >>"
        th.raise RuntimeError, mesg

        begin
            th.join
        rescue Exception => e
            assert_kind_of(RuntimeError, e)
            assert_equal(mesg, e.message)
        end
    end
end
